//
// Created by q1w2e3r4 on 23-8-13.
//

#include "HIR-opt/LoopPass.h"
#include <queue>
#include <vector>

vector<pair<BasicBlock*, BasicBlock*>> backedge; // <header,latch>
map<BasicBlock*,Loop*> belong_loop;
vector<Loop*> loop_info;

void Mark_depth(Function* func);
void mark(Loop * loop);
void build_lcssa(Loop* loop,Function* func);
void add_preheader(Loop* loop,Function* func);
void build_dedicated_exit(Loop* loop,Function* func);
void build_single_latch(Loop* loop,Function* func);
void Split_PHI(vector<BasicBlock*>& from, BasicBlock* to, BasicBlock* median);

ValueRef* get_correct_ref(ValueRef* ref);

void PostOrder(DomNode * dom){
    for(auto node: dom->child){
        PostOrder(node);
    }

    for(auto pred: dom->bb->pred){
        if(pred->domNode->nodeId > dom->nodeId){
            // find a ledge
            backedge.emplace_back(dom->bb,pred);
            //cerr << "Backedge: " << pred->label << " -> " << dom->bb->label << endl;
        }
    }


}

void LoopPass::analysis() {
    for(auto& [name,func]: gu->func_table){
        if(func->entry != nullptr) {
            backedge.clear();
            belong_loop.clear();
            loop_info.clear();
            FindLoop(func);
            Mark_depth(func);
        }
    }
}

void LoopPass::FindLoop(Function *func) {
    // step 1,find latchs
    PostOrder(func->entry->domNode);

    // step 2,find loop
    // one BasicBlock should have only one loop, if not ,merge them
    for(int i=0;i<backedge.size();++i){
        BasicBlock * header = backedge[i].first;
        Loop * loop = new Loop();
        loop->header = header;

        queue<BasicBlock *> q;
        q.push(backedge[i].second);
        loop->latch.push_back(backedge[i].second);
        while(i+1<backedge.size() && backedge[i+1].first == header){
            q.push(backedge[i+1].second);
            loop->latch.push_back(backedge[i+1].second);
            ++i;
        }

        while(!q.empty()){
            BasicBlock * now = q.front(); q.pop();
            Loop* subloop = nullptr;
            auto it = belong_loop.find(now);
            if(it != belong_loop.end()) subloop = it->second;

            if(subloop != nullptr){
                loop->add_loop(subloop);
                subloop->fa_loop = loop;
                assert(subloop->header != header);
                for(auto pred: subloop->header->pred){
                    if(!subloop->contains(pred))
                        q.push(pred);
                }
            }
            else{
                loop->add_block(now);
                if(now != header) {
                    for (auto pred: now->pred) {
                        if(!loop->contains(pred))
                            q.push(pred);
                    }
                }
            }
        }

        for(auto block: loop->contain_blocks){
            belong_loop[block] = loop;
        }
        //loop->debug();
        loop_info.push_back(loop);
    }


}

void LoopPass::Lcssa() {
    for(auto& [name,func]: gu->func_table){
        if(func->entry != nullptr) {
            backedge.clear();
            belong_loop.clear();
            loop_info.clear();
            FindLoop(func);
            BuildLoop(func);
        }
    }
}

void LoopPass::BuildLoop(Function* func) {
    for(auto loop: loop_info){
        if(loop->fa_loop != nullptr) continue;
        build_lcssa(loop,func);
    }
}

void build_lcssa(Loop* loop,Function* func){
    // change it to LCSSA
    // now we only add preheader to each loop
    // TODO: num of latchs -> 1;
    // TODO: dedicated exit
    add_preheader(loop,func);

    bool can_opt = true;
    for(auto subloop:loop->contain_loops){
        build_lcssa(subloop,func);
        if(!subloop->can_promoted) can_opt = false;
    }
    if(can_opt){
        if(loop->only_one_exit()){
            // dedicated exit, can be promoted;
            build_dedicated_exit(loop,func);
            build_single_latch(loop,func);
        }
        loop->can_promoted = true;
        return;
    }
    else{
        loop->can_promoted = false;
    }
}

void add_preheader(Loop* loop,Function* func){
    string name = "preheader";
    BasicBlock* preheader = new BasicBlock(name,func);
    for(auto prev: loop->header->pred){
        if(!loop->contains(prev))
            Insert_Block(prev,loop->header,preheader);
    }
    auto& v = func->block_list;
    for(auto it = v.begin(); it != v.end(); ++it){
        if(*it == loop->header){
            v.insert(it,preheader);
            break;
        }
    }

    loop->preheader = preheader;
}

void build_dedicated_exit(Loop* loop,Function* func){
    if(loop->exiting.size() == 1){
        loop->single_exit = *(loop->exiting.begin());
        return;
    }
    else{
        string name = "single_exit";
        BasicBlock* single_exit = new BasicBlock(name,func);
        func->block_list.push_back(single_exit);

        vector<BasicBlock*> exiting;
        for(auto bb:loop->exiting){
            exiting.push_back(bb);
        }
        Split_PHI(exiting,loop->dedicated_exit,single_exit);
        loop->single_exit = single_exit;
    }
}

void build_single_latch(Loop* loop,Function* func){
    if(loop->latch.size() == 1){
        loop->single_latch = loop->latch[0];
        return;
    }
    else{
        string name = "single_latch";
        BasicBlock* single_latch = new BasicBlock(name,func);
        func->block_list.push_back(single_latch);

        Split_PHI(loop->latch,loop->header,single_latch);
        loop->single_latch = single_latch;
    }
}

void Mark_depth(Function* func){
    for(auto block:func->block_list){ // init;
        block->loop_depth = 0;
        block->belong_loop = nullptr;
    }
    for(auto loop:loop_info){
        if(loop->fa_loop == nullptr){
            loop->depth = 1;
            mark(loop);
        }
    }
}

void mark(Loop * loop){
    for(auto block: loop->contain_blocks){
        block->belong_loop = loop;
        block->loop_depth = loop->depth;
        //cerr << block->label << " " << block->loop_depth << endl;
    }

    for(auto subloop: loop->contain_loops){
        subloop->depth = loop->depth + 1;
        mark(subloop);
    }
}

void Split_PHI(vector<BasicBlock*>& from, BasicBlock* to, BasicBlock* median){
    // upt succ/pred
    for(auto from_block: from){
        auto it1 = std::find(from_block->succ.begin(), from_block->succ.end(), to);
        *it1 = median;

        auto it2 = std::find(to->pred.begin(),to->pred.end(), from_block);
        to->pred.erase(it2);

        median->pred.push_back(from_block);
    }

    median->succ.push_back(to);
    to->pred.push_back(median);

    IRInstruction * br = new BrInstruction(to);
    median->local_instr.push_back(br);

    // modify PHI
    for(auto instr:to->local_instr){
        if(instr->instType == PHI){
            auto old_phi = dynamic_cast<PhiInstruction*>(instr);
            ValueRef* new_phi = get_correct_ref(*old_phi->def_list.begin());
            auto phi = new PhiInstruction(nullptr,new_phi,from.size());
            for(auto bb:from){
                ValueRef* val = old_phi->mp[bb];
                phi->mp[bb] = val;
                old_phi->mp.erase(bb); --old_phi->branch_cnt;

                // del use in ref
                auto it1 = std::find(val->use.begin(), val->use.end(), old_phi);
                *it1 = phi;

                // del use in instr
                old_phi->use_list.erase(val);
                old_phi->use_list.erase(bb);
            }
            Insert_instr_beforeBr(phi,median);

            // add use in instr/ref
            old_phi->mp[median] = new_phi;
            new_phi->use.push_back(old_phi);
            old_phi->use_list.insert(new_phi);
        }
    }


}

ValueRef* get_correct_ref(ValueRef* ref){
    if(ref->type == IntVar){
        return new Int_Var();
    }
    else if(ref->type == FloatVar){
        return new Float_Var();
    }
    else{
        cerr << "Invalid Type!" << endl;
        throw exception();
    }
}